'use strict';

class Deflower {
  constructor() {
    this._values = [];
    this._lines = [];
    this._line = [];
    this._remainder = null;
  }
}
module.exports = Deflower;
var sharedDeflower = new Deflower();

var Buffer = require('buffer').Buffer;
var cellflow = require(__dirname+'/cellflow');

/**
 * push() pushes a buffer to deflow
 * 
 * @param {Buffer} buffer - holds a buffer to deflow.
 * @return {Object} a count state of deflower.
 */
Deflower.prototype.push = function(buffer) {

  if (!Buffer.isBuffer(buffer)) {
    if (typeof(buffer) === 'string') {
      buffer = Buffer.from(buffer);
    } else {
      throw new TypeError('The input buffer argument is not a buffer');
    }
  }

  var remain = 0;
  if (this._remainder) {
    remain = this._remainder.length;
    buffer = Buffer.concat([this._remainder, buffer]);
    this._remainder = null;
  }

  var length = buffer.length;
  var offset = 0;

  while (offset < length) {

    // info
    var info = buffer.readUInt8(offset);

    // break
    if (info == 0xFF) {
      push(this, cellflow.breakSymbol);
      offset += 1;
      continue;
    }

    // null
    if (info === 0x00) {
      push(this, null);
      offset += 1;
      continue;
    }

    // boolean
    if (info === 0x10) {
      if (offset + 2 > length) break;

      let value = buffer.readUInt8(offset + 1);
      if (value !== 0 && value !== 1) {
        let error = new SyntaxError('A boolean cell expects 0 or 1 instead of '+ value);
        error.celltype = 'boolean';
        error.offset = offset - remain;
        return { error };
      }
      value = !!value;

      push(this, value);
      offset += 2;
      continue;
    }

    // integer
    if (info >= 0x01 && info <= 0x06) {
      if (offset + 1 + info > length) break;

      let value = buffer.readIntLE(offset + 1, info);

      push(this, value);
      offset += 1 + info;
      continue;
    }

    // float64
    if (info == 0x0A) {
      if (offset + 9 > length) break;

      let value = buffer.readDoubleLE(offset + 1);

      push(this, value);
      offset += 9;
      continue;
    }

    // string
    if (info >= 0x11 && info <= 0x14) {
      let sizeLength = info - 0x10;

      let start = offset + 1 + sizeLength;
      if (start > length) break;

      let size = buffer.readUIntLE(offset + 1, sizeLength);
      let end = offset + 1 + sizeLength + size;
      if (end > length) break;

      let value = buffer.toString('utf8', start, end);

      push(this, value);
      offset = end;
      continue;
    }

    // json
    if (info >= 0x21 && info <= 0x24) {
      let sizeLength = info - 0x20;

      let start = offset + 1 + sizeLength;
      if (start > length) break;

      let size = buffer.readUIntLE(offset + 1, sizeLength);
      let end = offset + 1 + sizeLength + size;
      if (end > length) break;

      let value = buffer.toString('utf8', start, end);
      try {
        value = JSON.parse(value);
      } catch (error) {
        error.celltype = 'json';
        error.offset = offset - remain;
        return { error };
      }

      push(this, value);
      offset = end;
      continue;
    }

    let error = new SyntaxError('Unexpected token 0x'+info.toString(16));
    error.celltype = 'unknown';
    error.offset = offset - remain;
    return { error };
  }

  if (offset < buffer.length) {
    this._remainder = Buffer.from(buffer.slice(offset, buffer.length));
  }

  return {
    values: this._values.length,
    lines: this._lines.length,
    remainders: this._remainder ? this._remainder.length : 0 
  };
};

/**
 * getValues() returns all avaible values, and empty cache.
 * 
 * @return {Array} an array of values
 */
Deflower.prototype.getValues = function() {
  var values = this._values;
  this._values = [];
  this._lines = [];
  this._line = [];
  return values;
};

/**
 * getLines() returns all avaible lines, and empty lines,
 * and remove from values cache which values were in line.
 * 
 * @return {Array} an array of lines.
 */
Deflower.prototype.getLines = function() {
  var lines = this._lines;
  this._values = this._line.slice(0);
  this._lines = [];
  return lines;
};

function push(deflower, value) {
  deflower._values.push(value);
  if (value === cellflow.breakSymbol) {
    deflower._lines.push(deflower._line);
    deflower._line = [];
  } else {
    deflower._line.push(value);
  }
}

/**
 * deflow() converts a buffer to an array of lines.
 * 
 * @return {Array} an array of lines.
 */
Deflower.deflow = function(buffer) {
  var deflower = new Deflower();
  var state = deflower.push(buffer);
  return  deflower.getLines();
};