'use strict';

class Enflower {
  constructor() {
    this._chunks = [];
  }
}
module.exports = Enflower;
var sharedEnflower = new Enflower();

var Buffer = require('buffer').Buffer;
var cellflow = require(__dirname+'/cellflow');

var nullBuffer = Buffer.from([0x00]);
var trueBuffer = Buffer.from([0x10, 0x01]);
var falseBuffer = Buffer.from([0x10, 0x00]);
var breakBuffer = Buffer.from([0xFF]);

/**
 * push() pushes a single value into flow.
 * 
 * @param {Any} value - holds a value in any supported type.
 */
Enflower.prototype.push = function(value) {

  if (value === undefined || value === null || value === NaN) {
    this._chunks.push(nullBuffer);
    return;
  }

  if (value === cellflow.breakSymbol) {
    this._chunks.push(breakBuffer);
    return;
  }

  let type = typeof(value);

  if (type === 'boolean') {
    this._chunks.push(value ? trueBuffer : falseBuffer);
    return;
  }

  if (type === 'number' && value === parseInt(value)) {
    let length = lengthOfInt(value);
    let buffer = Buffer.alloc(1 + length);

    buffer.writeUInt8(0x00 + length, 0);
    buffer.writeIntLE(value, 1, length);

    this._chunks.push(buffer);
    return;
  }

  if (type === 'number') {
    let buffer = Buffer.alloc(1 + 8);

    buffer.writeUInt8(0x0A, 0);
    buffer.writeDoubleLE(value, 1);
    
    this._chunks.push(buffer);
    return;
  }

  if (type === 'string') {
    let size = Buffer.byteLength(value);
    let length = lengthOfUInt(size);
    if (length > 4) {
      throw new TypeError('Cellflow supports string with length up to 2^32 only.');
    }
    let buffer = Buffer.alloc(1 + length + size);

    buffer.writeUInt8(0x10 + length, 0);
    buffer.writeUIntLE(size, 1, length);
    buffer.write(value, 1 + length);

    this._chunks.push(buffer);
    return;
  }

  if (type === 'object') {
    value = JSON.stringify(value);

    let size = Buffer.byteLength(value);
    let length = lengthOfUInt(size);
    let buffer = Buffer.alloc(1 + length + size);

    buffer.writeUInt8(0x20 + length, 0);
    buffer.writeUIntLE(size, 1, length);
    buffer.write(value, 1 + length);

    this._chunks.push(buffer);
    return;
  }

  throw new TypeError('Cellflow supports boolean, number, string and object only');

};

/**
 * break() pushes a break cell into the flow to break the flow line.
 */
Enflower.prototype.break = function() {
  this.push(cellflow.breakSymbol);
};

/**
 * line() pushes an array of values into the flow and breaks the line.
 * 
 * @param {Array} values - holds an array of values.
 */
Enflower.prototype.line = function(values) {
  for (let i = 0; i < values.length; i++) {
    this.push(values[i]);
  }
  this.break();
};

/**
 * get() retrieves all cell in flow in buffer
 * 
 * @return {Buffer} a buffer contains flow.
 */
Enflower.prototype.get = function() {
  var buffer = Buffer.concat(this._chunks);
  this._chunks = [];
  return buffer;
};

/**
 * enflow() converts an array of value to flow ending with break.
 * 
 * @return {Buffer} a buffer contains flow.
 */
Enflower.enflow = function(values) {
  for (let i = 0; i < values.length; i++) {
    sharedEnflower.push(values[i]);
  }
  sharedEnflower.break();
  return sharedEnflower.get();
};

function lengthOfInt(value) {
  value = value >= 0 ? value : -value;
  return (value <= 0x7F      ) ? 1
    : ((value <= 0x7FFF      ) ? 2
    : ((value <= 0x7FFFFF    ) ? 3
    : ((value <= 0x7FFFFFFF  ) ? 4
    : ((value <= 0x7FFFFFFFFF) ? 5
    : 6))));
}

function lengthOfUInt(value) {
  return (value <= 0xFF      ) ? 1
    : ((value <= 0xFFFF      ) ? 2
    : ((value <= 0xFFFFFF    ) ? 3
    : ((value <= 0xFFFFFFFF  ) ? 4
    : ((value <= 0xFFFFFFFFFF) ? 5
    : 6))));
}