var cellflow = require(__dirname+'/../');
var Deflower = cellflow.Deflower;

var Buffer = require('buffer').Buffer;
var assert = require('chai').assert;

describe('lib/deflower', () => {

  describe('#push #getValues', () => {

    it('should succesfully deflow a break', () => {
      var values = [cellflow.breakSymbol];
      var buffer = Buffer.from([0xFF]);
      var count = {
      	values: values.length,
      	lines: 1,
      	remainders: 0
      };

      var deflower = new Deflower();
      var state = deflower.push(buffer);
      assert.deepEqual(state, count);

      var result = deflower.getValues();
      assert.deepEqual(values, result);
      for (let i = 0; i < values.length; i++) {
        assert.strictEqual(values[i], result[i]);
      }
    });

    it('should succesfully deflow a null flow', () => {
      var values = [null];
      var buffer = Buffer.from([0x00]);
      var count = {
      	values: values.length,
      	lines: 0,
      	remainders: 0
      };

      var deflower = new Deflower();
      var state = deflower.push(buffer);
      assert.deepEqual(state, count);

      var result = deflower.getValues();
      assert.deepEqual(values, result);
      for (let i = 0; i < values.length; i++) {
        assert.strictEqual(values[i], result[i]);
      }
    });

    it('should succesfully deflow boolean values', () => {
      var values = [
        true,
        false,
        true,
        false
      ];
      var buffer = Buffer.from([
        0x10, 0x01,
        0x10, 0x00,
        0x10, 0x01,
        0x10, 0x00
      ]);      
      var count = {
      	values: values.length,
      	lines: 0,
      	remainders: 0
      };

      var deflower = new Deflower();
      var state = deflower.push(buffer);
      assert.deepEqual(state, count);

      var result = deflower.getValues();
      assert.deepEqual(values, result);
      for (let i = 0; i < values.length; i++) {
        assert.strictEqual(values[i], result[i]);
      }
    });

    it('should failed deflow a corrupted boolean value', () => {
      var values = [
        true,
      ];
      var buffer = Buffer.from([
        0x10, 0x02
      ]);

      var deflower = new Deflower();
      var state = deflower.push(buffer);
      assert.property(state, 'error');
      assert.instanceOf(state.error, SyntaxError);
      assert.equal(state.error.celltype, 'boolean');
      assert.equal(state.error.offset, 0);
    });

    it('should failed deflow boolean values', () => {
      var values = [
        true,
        false,
        true,
        false
      ];
      var buffer = Buffer.from([
        0x10, 0x01,
        0x10, 0x00,
        0x10, 0x01,
        0x10, 0x00
      ]);
      var count = {
      	values: values.length,
      	lines: 0,
      	remainders: 0
      };

      var deflower = new Deflower();
      var state = deflower.push(buffer);
      assert.deepEqual(state, count);

      var result = deflower.getValues();
      assert.deepEqual(values, result);
      for (let i = 0; i < values.length; i++) {
        assert.strictEqual(values[i], result[i]);
      }
    });

    it('should succesfully deflow boolean values, and have remainders', () => {
      var values = [
        true,
        false
      ];
      var buffer = Buffer.from([
        0x10, 0x01,
        0x10, 0x00,
        0x10,
      ]);
       var count = {
      	values: values.length,
      	lines: 0,
      	remainders: 1
      };

      var deflower = new Deflower();
      var state = deflower.push(buffer);
      assert.deepEqual(state, count);

      var result = deflower.getValues();
      assert.deepEqual(values, result);
      for (let i = 0; i < values.length; i++) {
        assert.strictEqual(values[i], result[i]);
      }
    });

    it('should succesfully deflow positive integer values less than 2^47 - 1', () => {
      var values = [
        0x00,
        0x71,
        0x7172,
        0x717273,
        0x71727374,
        0x7172737475,
        0x717273747576,
        0x7FFFFFFFFFFF
      ];
      var buffer = Buffer.from([
        0x01, 0x00,
        0x01, 0x71,
        0x02, 0x72, 0x71,
        0x03, 0x73, 0x72, 0x71,
        0x04, 0x74, 0x73, 0x72, 0x71,
        0x05, 0x75, 0x74, 0x73, 0x72, 0x71,
        0x06, 0x76, 0x75, 0x74, 0x73, 0x72, 0x71,
        0x06, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x7F,
      ]);
      var count = {
      	values: values.length,
      	lines: 0,
      	remainders: 0
      };

      var deflower = new Deflower();
      var state = deflower.push(buffer);
      assert.deepEqual(state, count);

      var result = deflower.getValues();
      assert.deepEqual(values, result);
      for (let i = 0; i < values.length; i++) {
        assert.strictEqual(values[i], result[i]);
      }
    });

    it('should succesfully deflow positive integer values and have remainders', () => {
      var values = [
        0x00,
        0x71,
      ];
      var buffer = Buffer.from([
        0x01, 0x00,
        0x01, 0x71,
        0x02, 0x72
      ]);
      var count = {
      	values: values.length,
      	lines: 0,
      	remainders: 2
      };

      var deflower = new Deflower();
      var state = deflower.push(buffer);
      assert.deepEqual(state, count);

      var result = deflower.getValues();
      assert.deepEqual(values, result);
      for (let i = 0; i < values.length; i++) {
        assert.strictEqual(values[i], result[i]);
      }
    });

    it('should succesfully deflow negative integer values more than -(2^47 - 1)', () => {
      var values = [
        -2,
      ];
      var buffer = Buffer.from([
        0x01, 0xFE,
      ]);
      var count = {
      	values: values.length,
      	lines: 0,
      	remainders: 0
      };

      var deflower = new Deflower();
      var state = deflower.push(buffer);
      assert.deepEqual(state, count);

      var result = deflower.getValues();
      assert.deepEqual(values, result);
      for (let i = 0; i < values.length; i++) {
        assert.strictEqual(values[i], result[i]);
      }
    });

    it('should succesfully deflow negative integer values and have remainders', () => {
      var values = [
        -2,
      ];
      var buffer = Buffer.from([
        0x01, 0xFE,
        0x02
      ]);
      var count = {
      	values: values.length,
      	lines: 0,
      	remainders: 1
      };

      var deflower = new Deflower();
      var state = deflower.push(buffer);
      assert.deepEqual(state, count);

      var result = deflower.getValues();
      assert.deepEqual(values, result);
      for (let i = 0; i < values.length; i++) {
        assert.strictEqual(values[i], result[i]);
      }
    });

    it('should succesfully deflow a float', () => {
      var values = [1.10];
      var buffer = Buffer.from([0x0A, 0x9A, 0x99, 0x99, 0x99, 0x99, 0x99, 0xF1, 0x3f]);
      var count = {
      	values: values.length,
      	lines: 0,
      	remainders: 0
      };

      var deflower = new Deflower();
      var state = deflower.push(buffer);
      assert.deepEqual(state, count);

      var result = deflower.getValues();
      assert.deepEqual(values, result);
      for (let i = 0; i < values.length; i++) {
        assert.strictEqual(values[i], result[i]);
      }
    });

    it('should succesfully deflow a float and have remainders', () => {
      var values = [1.10];
      var buffer = Buffer.from([
        0x0A, 0x9A, 0x99, 0x99, 0x99, 0x99, 0x99, 0xF1, 0x3F,
        0x0A, 0x9A, 0x99, 0x99, 0x99, 0x99, 0x99, 0xF1]);

      var count = {
      	values: values.length,
      	lines: 0,
      	remainders: 8
      };
      
      var deflower = new Deflower();
      var state = deflower.push(buffer);
      assert.deepEqual(state, count);

      var result = deflower.getValues();
      assert.deepEqual(values, result);
      for (let i = 0; i < values.length; i++) {
        assert.strictEqual(values[i], result[i]);
      }
    });

    it('should succesfully deflow a short string', () => {
      var values = ['short'];
      var buffer = Buffer.concat([
        Buffer.from([0x11, values[0].length]),
        Buffer.from(values[0])
      ]);

      var count = {
      	values: values.length,
      	lines: 0,
      	remainders: 0
      };
      var deflower = new Deflower();
      var state = deflower.push(buffer);
      assert.deepEqual(state, count);

      var result = deflower.getValues();
      assert.deepEqual(values, result);
      for (let i = 0; i < values.length; i++) {
        assert.strictEqual(values[i], result[i]);
      }
    });

    it('should succesfully deflow a short string and have remainders of size', () => {
      var values = ['short'];
      var buffer = Buffer.concat([
        Buffer.from([0x11, values[0].length]),
        Buffer.from(values[0]),
        Buffer.from([0x11])
      ]);

      var count = {
      	values: values.length,
      	lines: 0,
      	remainders: 1
      };
      
      var deflower = new Deflower();
      var state = deflower.push(buffer);
      assert.deepEqual(state, count);

      var result = deflower.getValues();
      assert.deepEqual(values, result);
      for (let i = 0; i < values.length; i++) {
        assert.strictEqual(values[i], result[i]);
      }
    });

    it('should succesfully deflow a short string and have remainders of data', () => {
      var values = ['short'];
      var buffer = Buffer.concat([
        Buffer.from([0x11, values[0].length]),
        Buffer.from(values[0]),
        Buffer.from([0x11, 0x02])
      ]);

      var count = {
      	values: values.length,
      	lines: 0,
      	remainders: 2
      };
      
      var deflower = new Deflower();
      var state = deflower.push(buffer);
      assert.deepEqual(state, count);

      var result = deflower.getValues();
      assert.deepEqual(values, result);
      for (let i = 0; i < values.length; i++) {
        assert.strictEqual(values[i], result[i]);
      }
    });

    it('should succesfully deflow a long string', () => {
      var values = [
        '0123456789'+'0123456789'+'0123456789'+'0123456789'+'0123456789'+
        '0123456789'+'0123456789'+'0123456789'+'0123456789'+'0123456789'+
        '0123456789'+'0123456789'+'0123456789'+'0123456789'+'0123456789'+
        '0123456789'+'0123456789'+'0123456789'+'0123456789'+'0123456789'+
        '0123456789'+'0123456789'+'0123456789'+'0123456789'+'0123456789'+
        '0123456789'+'0123456789'+'0123456789'+'0123456789'+'0123456789'
      ];
      var buffer = Buffer.concat([
        Buffer.from([0x12, values[0].length & 0xFF, values[0].length >> 8]),
        Buffer.from(values[0])
      ]);
      var count = {
      	values: values.length,
      	lines: 0,
      	remainders: 0
      };

      var deflower = new Deflower();
      var state = deflower.push(buffer);
      assert.deepEqual(state, count);

      var result = deflower.getValues();
      assert.deepEqual(values, result);
      for (let i = 0; i < values.length; i++) {
        assert.strictEqual(values[i], result[i]);
      }  
    });

    it('should succesfully deflow an object', () => {
      var values = [{'number':1,array:[3]}];
      var jstr = JSON.stringify(values[0]);
      var buffer = Buffer.concat([
        Buffer.from([0x21, jstr.length]),
        Buffer.from(jstr)
      ]);
      var count = {
      	values: values.length,
      	lines: 0,
      	remainders: 0
      };
      
      var deflower = new Deflower();
      var state = deflower.push(buffer);
      assert.deepEqual(state, count);

      var result = deflower.getValues();
      assert.deepEqual(values, result);
      for (let i = 0; i < values.length; i++) {
        assert.deepEqual(values[i], result[i]);
      }
    });

    it('should succesfully deflow an object and have remainders of size', () => {
      var values = [{'number':1,array:[3]}];
      var jstr = JSON.stringify(values[0]);
      var buffer = Buffer.concat([
        Buffer.from([0x21, jstr.length]),
        Buffer.from(jstr),
        Buffer.from([0x22, 0x01])
      ]);
      var count = {
      	values: values.length,
      	lines: 0,
      	remainders: 2
      };
      
      var deflower = new Deflower();
      var state = deflower.push(buffer);
      assert.deepEqual(state, count);

      var result = deflower.getValues();
      assert.deepEqual(values, result);
      for (let i = 0; i < values.length; i++) {
        assert.deepEqual(values[i], result[i]);
      }
    });

    it('should succesfully deflow an object and have remainders of data', () => {
      var values = [{'number':1,array:[3]}];
      var jstr = JSON.stringify(values[0]);
      var buffer = Buffer.concat([
        Buffer.from([0x21, jstr.length]),
        Buffer.from(jstr),
        Buffer.from([0x21, 0x01])
      ]);
      var count = {
      	values: values.length,
      	lines: 0,
      	remainders: 2
      };
      
      var deflower = new Deflower();
      var state = deflower.push(buffer);
      assert.deepEqual(state, count);

      var result = deflower.getValues();
      assert.deepEqual(values, result);
      for (let i = 0; i < values.length; i++) {
        assert.deepEqual(values[i], result[i]);
      }
    });

    it('should failed to deflow a corrupted object', () => {
      var values = [{'number':1,array:[3]}];
      var jstr = JSON.stringify(values[0]).replace('{', '[');
      var buffer = Buffer.concat([
        Buffer.from([0x21, jstr.length]),
        Buffer.from(jstr)
      ]);
      
      var deflower = new Deflower();
      var state = deflower.push(buffer);
      assert.property(state, 'error');
      assert.instanceOf(state.error, SyntaxError);
      assert.equal(state.error.celltype, 'json');
      assert.equal(state.error.offset, 0);
    });

    it('should failed deflow an invalid cell', () => {
      var buffer = Buffer.from([
        0x30, 0x01
      ]);

      var deflower = new Deflower();
      var state = deflower.push(buffer);
      assert.property(state, 'error');
      assert.instanceOf(state.error, SyntaxError);
      assert.equal(state.error.celltype, 'unknown');
      assert.equal(state.error.offset, 0);
    });

    it('should succesfully deflow with combining remainders', () => {
      var values = [
        0x00,
        0x71,
        0x72
      ];
      var buffer = Buffer.from([
        0x01, 0x00,
        0x01, 0x71,
        0x01
      ]);
      var nextBuffer = Buffer.from([
        0x72
      ]);
      var count = {
      	values: values.length - 1,
      	lines: 0,
      	remainders: 1
      };
      var nextCount = {
      	values: values.length,
      	lines: 0,
      	remainders: 0
      };

      var deflower = new Deflower();
      var state = deflower.push(buffer);
      assert.deepEqual(state, count);
      var nextState = deflower.push(nextBuffer);
      assert.deepEqual(nextState, nextCount);

      var result = deflower.getValues();
      assert.deepEqual(values, result);
      for (let i = 0; i < values.length; i++) {
        assert.strictEqual(values[i], result[i]);
      }
    });

  });

  describe('#break', () => {

    it('should succesfully deflow a single break', () => {
      var values = [cellflow.breakSymbol];
      var buffer = Buffer.from([0xFF]);
      var count = {
      	values: values.length,
      	lines: 1,
      	remainders: 0
      };

      var deflower = new Deflower();
      var state = deflower.push(buffer);
      assert.deepEqual(state, count);

      var result = deflower.getValues();
      assert.deepEqual(values, result);
      for (let i = 0; i < values.length; i++) {
        assert.strictEqual(values[i], result[i]);
      }
    });

    it('should succesfully deflow double breaks', () => {
      var values = [cellflow.breakSymbol, cellflow.breakSymbol];
      var buffer = Buffer.from([0xFF, 0xFF]);
      var count = {
      	values: values.length,
      	lines: 2,
      	remainders: 0
      };

      var deflower = new Deflower();
      var state = deflower.push(buffer);
      assert.deepEqual(state, count);

      var result = deflower.getValues();
      assert.deepEqual(values, result);
      for (let i = 0; i < values.length; i++) {
        assert.strictEqual(values[i], result[i]);
      }
    });

    it('should succesfully deflow a null flow followed by break', () => {
      var values = [null, cellflow.breakSymbol];
      var buffer = Buffer.from([0x00, 0xFF]);
      var count = {
      	values: values.length,
      	lines: 1,
      	remainders: 0
      };

      var deflower = new Deflower();
      var state = deflower.push(buffer);
      assert.deepEqual(state, count);

      var result = deflower.getValues();
      assert.deepEqual(values, result);
      for (let i = 0; i < values.length; i++) {
        assert.strictEqual(values[i], result[i]);
      }
    });

    it('should succesfully deflow a null flow followed by 2 breaks', () => {
      var values = [null, cellflow.breakSymbol, cellflow.breakSymbol];
      var buffer = Buffer.from([0x00, 0xFF, 0xFF]);
      var count = {
      	values: values.length,
      	lines: 2,
      	remainders: 0
      };

      var deflower = new Deflower();
      var state = deflower.push(buffer);
      assert.deepEqual(state, count);

      var result = deflower.getValues();
      assert.deepEqual(values, result);
      for (let i = 0; i < values.length; i++) {
        assert.strictEqual(values[i], result[i]);
      }
    });

  });

  describe('#getLines', () => {

    it('should get 0 line after deflowing a booleans', () => {
      var lines = [];
      var buffer = Buffer.from([0x10, 0x01, 0x10, 0x00]);
      var count = {
      	values: 2,
      	lines: 0,
      	remainders: 0
      };

      var deflower = new Deflower();
      var state = deflower.push(buffer);
      assert.deepEqual(state, count);

      var result = deflower.getLines();
      assert.deepEqual(lines, result);
      for (let i = 0; i < lines.length; i++) {
        assert.strictEqual(lines[i], result[i]);
      }
    });

    it('should get 1 line after deflowing booleans with a break among them', () => {
      var lines = [[true]];
      var buffer = Buffer.from([0x10, 0x01, 0xFF, 0x10, 0x00]);
      var count = {
      	values: 3,
      	lines: 1,
      	remainders: 0
      };

      var deflower = new Deflower();
      var state = deflower.push(buffer);
      assert.deepEqual(state, count);

      var result = deflower.getLines();
      assert.deepEqual(lines, result);
      for (let i = 0; i < lines.length; i++) {
        assert.deepEqual(lines[i], result[i]);
      }
    });

  });

  describe('#deflow', () => {

    it('should succesfully deflow a buffer into an array of lines', () => {
      var lines = [[
        null,
        null,
        true,
        false,
        1,
        -1
      ], []];
      var buffer = Buffer.from([
        0x00,
        0x00,
        0x10, 0x01,
        0x10, 0x00,
        0x01, 0x01,
        0x01, 0xFF,
        0xFF,
        0xFF
      ]);

      var result = Deflower.deflow(buffer);
      assert.deepEqual(result, lines);  
    });

  });

});