var Enflower = require(__dirname+'/../').Enflower;

var Buffer = require('buffer').Buffer;
var assert = require('chai').assert;

describe('lib/enflower', () => {

  describe('#push', () => {

    it('should succesfully enflow a null value', () => {
      var value = null;
      var buffer = Buffer.from([0x00]);

      var enflower = new Enflower();

      enflower.push(value);
      var result = enflower.get();
      assert.equal(Buffer.compare(result, buffer), 0);
    });

    it('should succesfully enflow a undefined value', () => {
      var value = undefined;
      var buffer = Buffer.from([0x00]);

      var enflower = new Enflower();

      enflower.push(value);
      var result = enflower.get();
      assert.equal(Buffer.compare(result, buffer), 0);
    });

    it('should succesfully enflow a null, undefined values', () => {
      var values = [
        null,
        undefined
      ];
      var buffer = Buffer.from([
        0x00,
        0x00
      ]);

      var enflower = new Enflower();

      values.forEach((value) => {
        enflower.push(value);
      });
      var result = enflower.get();
      assert.equal(Buffer.compare(result, buffer), 0);
    });

    it('should succesfully enflow boolean values', () => {
      var values = [
        true,
        false,
        true,
        false
      ];
      var buffer = Buffer.from([
        0x10, 0x01,
        0x10, 0x00,
        0x10, 0x01,
        0x10, 0x00
      ]);

      var enflower = new Enflower();

      values.forEach((value) => {
        enflower.push(value);
      });
      var result = enflower.get();
      assert.equal(Buffer.compare(result, buffer), 0);    
    });

    it('should succesfully enflow positive integer values less than 2^47 - 1', () => {
      var values = [
        0x00,
        0x71,
        0x7172,
        0x717273,
        0x71727374,
        0x7172737475,
        0x717273747576,
        0x7FFFFFFFFFFF
      ];
      var buffer = Buffer.from([
        0x01, 0x00,
        0x01, 0x71,
        0x02, 0x72, 0x71,
        0x03, 0x73, 0x72, 0x71,
        0x04, 0x74, 0x73, 0x72, 0x71,
        0x05, 0x75, 0x74, 0x73, 0x72, 0x71,
        0x06, 0x76, 0x75, 0x74, 0x73, 0x72, 0x71,
        0x06, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x7F,
        ]);

      var enflower = new Enflower();

      values.forEach((value) => {
        enflower.push(value);
      });
      var result = enflower.get();
      assert.equal(Buffer.compare(result, buffer), 0);
    });

    it('should fail to enflow positive integer value more than 2^47 - 1', () => {
      var enflower = new Enflower();

      try {
        enflower.push(0x80FFFFFFFFFF);
        assert(false, 'should failed push');
      } catch (error) {
        assert.instanceOf(error, TypeError);
      }
    });

    it('should succesfully enflow negative integer values more than -(2^47 - 1)', () => {
      var values = [
        -2,
      ];
      var buffer = Buffer.from([
        0x01, 0xFE,
      ]);

      var enflower = new Enflower();

      values.forEach((value) => {
        enflower.push(value);
      });
      var result = enflower.get();
      assert.equal(Buffer.compare(result, buffer), 0);  
    });

    it('should succesfully enflow a float', () => {
       var value = 1.10;
      var buffer = Buffer.from([0x0A, 0x9A, 0x99, 0x99, 0x99, 0x99, 0x99, 0xF1, 0x3f]);

      var enflower = new Enflower();

      enflower.push(value);
      var result = enflower.get();
      assert.equal(Buffer.compare(result, buffer), 0);  
    });

    it('should succesfully enflow a short string', () => {
      var value = 'short';
      var buffer = Buffer.concat([
        Buffer.from([0x11, value.length]),
        Buffer.from(value)
      ]);

      var enflower = new Enflower();

      enflower.push(value);
      var result = enflower.get();
      assert.equal(Buffer.compare(result, buffer), 0); 
    });

    it('should succesfully enflow a long string', () => {
      var value = 
        '0123456789'+'0123456789'+'0123456789'+'0123456789'+'0123456789'+
        '0123456789'+'0123456789'+'0123456789'+'0123456789'+'0123456789'+
        '0123456789'+'0123456789'+'0123456789'+'0123456789'+'0123456789'+
        '0123456789'+'0123456789'+'0123456789'+'0123456789'+'0123456789'+
        '0123456789'+'0123456789'+'0123456789'+'0123456789'+'0123456789'+
        '0123456789'+'0123456789'+'0123456789'+'0123456789'+'0123456789'
        ;
      var buffer = Buffer.concat([
        Buffer.from([0x12, value.length & 0xFF, value.length >> 8]),
        Buffer.from(value)
      ]);

      var enflower = new Enflower();

      enflower.push(value);
      var result = enflower.get();
      assert.equal(Buffer.compare(result, buffer), 0);    
    });

    it('should succesfully enflow an object', () => {
      var value = {'number':1,array:[3]};
      var jstr = JSON.stringify(value);
      var buffer = Buffer.concat([
        Buffer.from([0x21, jstr.length]),
        Buffer.from(jstr)
      ]);

      var enflower = new Enflower();

      enflower.push(value);
      var result = enflower.get();
      assert.equal(Buffer.compare(result, buffer), 0);  
    });

    it('should fail to enflow a Symbol', () => {
      var enflower = new Enflower();

      try {
        enflower.push(Symbol('test'));
        assert(false, 'should failed push');
      } catch (error) {
        assert.instanceOf(error, TypeError);
      }
    });

  });

  describe('#break', () => {

    it('should succesfully enflow values and end with break', () => {
      var values = [
        null,
        undefined,
        true,
        false,
        1,
        -1
      ];
      var buffer = Buffer.from([
        0x00,
        0x00,
        0x10, 0x01,
        0x10, 0x00,
        0x01, 0x01,
        0x01, 0xFF,
        0xFF
      ]);

      var enflower = new Enflower();

      values.forEach((value) => {
        enflower.push(value);
      });
      enflower.break();
      var result = enflower.get();
      assert.equal(Buffer.compare(result, buffer), 0);   
    });

  });

  describe('#line', () => {

    it('should succesfully enflow values and end with break', () => {
      var values = [
        null,
        undefined,
        true,
        false,
        1,
        -1
      ];
      var buffer = Buffer.from([
        0x00,
        0x00,
        0x10, 0x01,
        0x10, 0x00,
        0x01, 0x01,
        0x01, 0xFF,
        0xFF
      ]);

      var enflower = new Enflower();

      enflower.line(values);
      var result = enflower.get();
      assert.equal(Buffer.compare(result, buffer), 0);  
    });

  });

  describe('#enflow', () => {

    it('should succesfully enflow values and the result contains break', () => {
      var values = [
        null,
        undefined,
        true,
        false,
        1,
        -1
      ];
      var buffer = Buffer.from([
        0x00,
        0x00,
        0x10, 0x01,
        0x10, 0x00,
        0x01, 0x01,
        0x01, 0xFF,
        0xFF
      ]);

      var result = Enflower.enflow(values);
      assert.equal(Buffer.compare(result, buffer), 0);  
    });

  });

});